
import std.stdio;
import std.format;
import core.sys.posix.unistd;
import core.sys.linux.epoll;
import msgq;
import esrc;

align (1) struct EpollEvent {
    align(1):
    uint event_mask;
    EventSource es;
    /* just do not want to use that union, epoll_data_t */
}
static assert(EpollEvent.sizeof == 12);

extern (C) {
    int epoll_ctl(int epfd, int op, int fd, EpollEvent *desiredEvents);
    int epoll_wait(int epfd, EpollEvent *events, int maxevents, int timeout);
}

final class EventQueue {

    enum uint ePollIn = EPOLLIN;
    enum uint ePollOut = EPOLLOUT;
    enum uint ePollErr = EPOLLERR;
    enum uint ePollHup = EPOLLHUP;
    enum uint ePollRdHup = EPOLLRDHUP;

    static private EventQueue _eq;

    private int id;
    private bool done;
    private MessageQueue mq;

    private this() {
        id = epoll_create1(0);
        assert(id > 0, "epoll_create1() failed");
        mq = MessageQueue.getInstance();
    }

    private ~this() {
        writeln(__FUNCTION__);
        close(id);
    }

    static EventQueue getInstance() {
        if (_eq is null)
            _eq = new EventQueue();
        return _eq;
    }

    void registerEventSource(EventSource es) {
        auto e = EpollEvent(0, es);
        int r = epoll_ctl(id, EPOLL_CTL_ADD, es.id, &e);
        assert(r == 0, "epoll_ctl(ADD) failed");
    }

    EventQueue opOpAssign(string op)(EventSource es)
    if (("+" == op) || ("~" == op)) {
        registerEventSource(es);
        return this;
    }

    void deregisterEventSource(EventSource es) {
        auto e = EpollEvent(0, es);
        int r = epoll_ctl(id, EPOLL_CTL_DEL, es.id, &e);
        assert(r == 0, "epoll_ctl(DEL) failed");
    }

    EventQueue opOpAssign(string op)(EventSource es)
    if ("-" == op) {
        deregisterEventSource(es);
        return this;
    }

    void enableEventSource(EventSource es, bool forCanWrite = false) {
        int em = forCanWrite ? EPOLLOUT : EPOLLIN;
        auto e = EpollEvent(em, es);
        int r = epoll_ctl(id, EPOLL_CTL_MOD, es.id, &e);
        assert(r == 0, "epoll_ctl(MOD) failed");
    }

    void enableEventSourceOnce(EventSource es, bool forCanWrite = false) {
        int em = forCanWrite ? EPOLLOUT : EPOLLIN;
        em |= EPOLLONESHOT;
        auto e = EpollEvent(em, es);
        int r = epoll_ctl(id, EPOLL_CTL_MOD, es.id, &e);
        assert(r == 0, "epoll_ctl(MOD) failed");
    }

    void disableEventSource(EventSource es) const {
        auto e = EpollEvent(0, es);
        int r = epoll_ctl(id, EPOLL_CTL_MOD, es.id, &e);
        assert(r == 0, "epoll_ctl(MOD) failed");
    }

    bool wait() {

        const int maxEvents = 8;
        EpollEvent[maxEvents] events;

        if (done)
            return false;

        int n = epoll_wait(id, events.ptr, maxEvents, -1);
        if (-1 == n)
            return false;

        foreach (k; 0 .. n) {
            EventSource s = events[k].es;
            ulong ecode = s.eventCode(events[k].event_mask);
            //mq.putMsg(null, s.owner, ecode, s);
            mq ~= Message(null, s.owner, ecode, s);
        }

        return true;
    }

    void stop() {
        done = true;
    }
}
