
import std.stdio;
import std.format;
import msgq;
import ecap;
import esrc;

alias Action = void delegate(StageMachine src, Object o);
alias Enter = void delegate();
alias Leave = void delegate();

final class Reflex {
    Action action;
    Stage nextStage;
}

final class Stage {
    string name;
    private Enter enter;
    private Leave leave;
    private Reflex[string] reflexes;

    final void addReflex(string eventName, Action a) {
        assert(
            eventName !in reflexes,
            format("'%s' already has reflex for '%s'", name, eventName)
        );

        Reflex r = new Reflex();
        r.action = a;
        reflexes[eventName] = r;
    }

    final void addReflex(string eventName, Stage s) {
        assert(
            eventName !in reflexes,
            format("'%s' already has reflex for '%s'", name, eventName)
        );

        Reflex r = new Reflex();
        r.nextStage = s;
        reflexes[eventName] = r;
    }
}

class StageMachine {
    string name;
    private bool isRunning;
    private Stage[string] stages;
    private Stage currentStage;

    private bool hasIo;
    private int tm_number;
    private int sg_number;

    private EventQueue eq;
    private MessageQueue mq;

    this(string name) {
        this.name = name;
        eq = EventQueue.getInstance();
        mq = MessageQueue.getInstance();
    }

/*
    ~this () {
        writefln("___!!!___%s(): %s destroyed...", __FUNCTION__, name);
    }
*/
    final Stage addStage(string stageName, Enter enter = null, Leave leave = null) {
        assert(
            stageName !in stages,
            format("'%s' already has stage '%s'", name, stageName)
        );

        Stage s = new Stage();
        s.name = stageName; // a couple of words about 'private'...
        s.enter = enter;
        s.leave = leave;
        stages[stageName] = s;
        return s;
    }

    final Timer newTimer() {
        Timer tm = new Timer();
        tm.owner = this;
        tm.number = tm_number++;

        tm.register();
        tm.enable();
        tm.stop();
        return tm;
    }

    final Signal newSignal(int signum) {
        Signal sg = new Signal(signum);
        sg.owner = this;
        sg.number = sg_number++;
        sg.register();
        return sg;
    }

    final Io newIo(int fd) {
        assert(!hasIo, format("'%s' already has Io channel", name));

        Io io = new Io(fd);
        io.owner = this;
        io.register();
        hasIo = true;
        return io;
    }

    final Io newIo() {
        assert(!hasIo, format("'%s' already has Io channel", name));

        Io io = new Io();
        io.owner = this;
        hasIo = true;
        return io;
    }

    final TCPListener newTCPListener(ushort port) {
        assert(!hasIo, format("'%s' already has Io channel", name));

        TCPListener l = new TCPListener(port);
        l.owner = this;
        hasIo = true;
        return l;
    }

    final ClientSocket newClientSocket(string host, string port) {
        assert(!hasIo, format("'%s' already has Io channel", name));

        ClientSocket s = new ClientSocket(host, port);
        s.owner = this;
        hasIo = true;
        return s;
    }

    final void run() {
        assert(!isRunning, format("'%s' is already started", name));
        assert(stages.length > 0, format("'%s' has no stages", name));

        Stage si = stages.get("INIT", null);
        assert(si !is null, format("'%s' has no 'INIT' stage", name));

        foreach (skey; stages.byKey) {
            Stage s = stages[skey];
            assert(s.reflexes.length > 0, format("'%s' has no reflexes for stage '%s'", name, s.name));
            foreach (rkey; s.reflexes.byKey) {
                Reflex r = s.reflexes[rkey];
                assert(
                    (r.action is null) ^ (r.nextStage is null),
                    format("'%s' has illegal reflex for '%s' at stage '%s'", name, rkey, s.name)
                );
            }
        }

        currentStage = si;
        si.enter();
        isRunning = true;
    }

    /* this method is state machine engine */
    final void reactTo(ref Message m) {
        string eventName;

        if (m.src is null) {
            EventSource ch = cast(EventSource)m.o;
            eventName = format("%c%u", ch.tag, m.code);
        } else {
            eventName = format("M%u", m.code);
        }

        writefln (
            "'%s @ %s' got '%s' from '%s'", name, currentStage.name, eventName,
            m.src ? (m.src is this ? "SELF" : m.src.name) : "OS"
        );

        if (eventName !in currentStage.reflexes) {

            writefln (
                "'%s @ %s' DROPPED '%s' from '%s'", name, currentStage.name, eventName,
                m.src ? (m.src is this ? "SELF" : m.src.name)  : "OS"
            );
            return;
        }

        Reflex r = currentStage.reflexes[eventName];

        /* action */
        if (r.action) {
            r.action(m.src, m.o);
            return;
        }

        /* transition */
        if (currentStage.leave)
            currentStage.leave();

        Stage nextStage = r.nextStage;

        if (nextStage.enter)
            nextStage.enter();

        currentStage = nextStage;
    }

    /*
    ** a machine can send a message to other machines by two ways:
    ** - msgTo(recipient, code, object);
    ** - recipient(this, code, object);
    */
    final void msgTo(StageMachine sm, ulong code, Object o = null) {
        mq.putMsg(this, sm, code, o);
        // as an alternative you can use '+=' or '~=' operators:
        // mq += Message(this, sm, code, o); // класссс...
        // mq ~= Message(this, sm, code, o);
    }

    final void opCall(StageMachine caller, ulong code, Object o = null) {
        mq.putMsg(caller, this, code, o);
    }

    /* да, мы знаем толк в извращениях :) */
    final ref StageMachine opUnary(string op)() if ("+" == op) {
        mq.putMsg(this, this, 0, null);
        return this;
    }

    final ref StageMachine opOpAssign(string op)(ulong mcode) if ("+" == op) {
        mq.putMsg(this, this, mcode, null);
        return this;
    }

    final void stopEventQueue() {
        eq.stop();
    }
}
